##############################################################################
#
# Copyright (c) 2001 Zope Corporation and Contributors. All Rights Reserved.
# 
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
# 
##############################################################################

import threading, time, sys, Queue, asyncore, socket
import traceback, re
from websiteloadtest.common.Request import Request
from websiteloadtest.xmlrpclib import xmlrpclib
from websiteloadtest.client.LogUDPClient import Logger
from websiteloadtest.common import Globals
from random import randrange

CLOSABLE = Queue.Queue()
numClosers = 20

MAX_SOCKET_LEN = Globals.CONFIG.getint('client','socketmaplength')

__version__ = "$Id: ClientReplayer.py,v 1.13 2002/02/25 18:49:07 andreas Exp $"

# regex to extract HTTP status code from HTTP reply
reg_http_hd = re.compile('HTTP/(.*) (.*) (.*)')

# STATS keeps HTTP related statistics
# keys are HTTP response codes (as string)
# values are the number of requests

STATS = {}
from asyncore import EINPROGRESS, EALREADY, EWOULDBLOCK,ECONNRESET,ESHUTDOWN,ENOTCONN

class ClientReplayer(threading.Thread):
    """ replayer to replay requests.
        Request are fetch in blocks of multiple requests
        using XMLRPC from the rp_server
    """

    running   = 1
    suspended = 1
    socket_closer = {}


    def __init__(self,rp_server=None,name='replayer'):

        threading.Thread.__init__(self)
        self.setName(name)

        self.rp_server = rp_server

        self.stats_logger = \
            Logger(Globals.STAT_LOG_SERVER_HOST, \
                   int(Globals.STAT_LOG_SERVER_PORT))

        # Startup socket closer threads

#        for i in range(numClosers):
#            self.socket_closer[i] = SocketCloser()
#            self.socket_closer[i].start()
#
        self.asyncore_thread = AsyncoreThread()
        self.asyncore_thread.start()


    def isSuspended(self):  return self.suspended
    def stop(self):         self.running   = 0

    def resume(self):       
        self.suspended = 0
        self.test_init()

    def suspend(self):      
        self.suspended = 1
        self.test_log_resume()


    def test_log_resume(self):

        duration  = time.time() - self.startTest

        s = '\tTest results:\n\n'
        s+= '\tRequests: %d, Errors: %d\n' % (self.numRequests,self.numErrors)
        s+= '\tElapsed time: %5.1f secs, Hits/sec: %5.2f\n' % \
                (duration, (1.0*self.numRequests/duration))

        keys = STATS.keys()
        keys.sort()

        s+='\tHTTP Stats:\n'

        for k in keys:
            s += '\t\t%3s: %d\n' % (k, STATS[k])

        self.stats_logger.log_info(s)



    def test_init(self):

        self.numRequests = 0
        self.numErrors   = 0
        self.startTest   = time.time()
        self.endTest     = 0 

        STATS.clear()

        Globals.LOG.log_info("replaying started")


    def run(self):

        secs_behind = 0

        while self.running:

            time.sleep(0.01)

            if not self.isSuspended():

                if secs_behind > 10:
                    Globals.LOG.log_warn("replayer is lagging behind: %5.2f" % (secs_behind)) 


                try:        
                    res = self.rp_server.getData()
                except:
                    traceback.print_exc()
                    continue

                if res['status'] == 1:

                    data = res['msg']

                    for d in data:
                        request = eval(d)
                        
                        async = AsyncRequest(request)

                        while request.timestamp > time.time() - self.startTest:
                            time.sleep(0.2)

                        secs_behind = time.time() - self.startTest - request.timestamp

                        try:
                            async.go()
                            self.numRequests+=1
                        except:
                            traceback.print_exc()
                            self.numErrors+=1

                        

                elif res['status'] == 0:
                    if res['msg'] == 'Feeder suspended':
                        Globals.LOG.log_info("Feeder is suspended, so let's suspend too")
                        self.suspend()
                        
                        
        Globals.LOG.log_info("%s terminated" % self.getName()) 


NUM_CLOSED = 0

class SocketCloser(threading.Thread):

    running = 1
    
    def __init__(self):

        threading.Thread.__init__(self)
        self.setName("socketcloser")


    def run(self):

        while self.running:

            try:
                socket, entered, delay = CLOSABLE.get(block=0)
                wait = delay - (time.time() - entered)
#                if wait > 0:
#                    time.sleep(wait)
                socket.close()
                global NUM_CLOSED
                NUM_CLOSED = NUM_CLOSED + 1
#                print NUM_CLOSED

            except Queue.Empty: # avoid deadlock
                time.sleep(.01)


        Globals.LOG.log_info("%s terminated" % self.getName(), p=1)  

    def stop(self): self.running = 0




class AsyncRequest:

    connected = 0

    def __init__(self, req ):

#        asyncore.dispatcher.__init__(self)

        self.delay     = req.throttle

        if Globals.PROXY_HOST:
            self.host      = Globals.PROXY_HOST
            self.port      = Globals.PROXY_PORT
            self.data      = req.getBody(fullurl=1)

        else:
            self.host      = req.host
            self.port      = int(req.port)
            self.data      = req.getBody(fullurl=0)


        self.req_host      = req.host        
        self.req_port      = req.port
        self.req_path      = req.path

    def send (self, data):
        try:
            result = self.socket.send (data)
            return result
        except socket.error, why:
            if why[0] == EWOULDBLOCK:
                return 0
            else:
                raise socket.error, why
            return 0

    def recv (self, buffer_size):
        try:
            data = self.socket.recv (buffer_size)
            if not data:
                # a closed connection is indicated by signaling
                # a read condition, and having recv() return 0.
                self.handle_close()
                return ''
            else:
                return data
        except socket.error, why:
            # winsock sometimes throws ENOTCONN
            if why[0] in [ECONNRESET, ENOTCONN, ESHUTDOWN]:
                self.handle_close()
                return ''
            else:
                raise socket.error, why
        
    def connect (self, address):
        self.connected = 0
        try:
            self.socket.connect (address)
        except socket.error, why:
            if why[0] in (EINPROGRESS, EALREADY, EWOULDBLOCK):
                return
            else:
                raise socket.error, why
        self.connected = 1
        self.handle_connect()

    def go(self):

#        print len(asyncore.socket_map)

        while len(asyncore.socket_map)>MAX_SOCKET_LEN:
            time.sleep(0.01)

        self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
        # self.connect sets self.connected properly
        try:

            proxy = '192.168.23.22'
            port = 9000

            self.connect((proxy, port))
#            self.connect((self.host, self.port))
        except:
            print 'not connected'
            traceback.print_exc()
            self.del_channel()
            self.socket.close()

    def create_socket (self, family, type):
        self.family_and_type = family, type
        self.socket = socket.socket (family, type)
        # 64K receive buffer
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, 65536)
        # 16K send buffer
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF, 16384)
        # do nonblocking IO
        self.socket.setblocking(0)
        self._fileno = self.socket.fileno()
        asyncore.socket_map[self._fileno] = self


    def del_channel (self):
        fd = self._fileno

        if asyncore.socket_map.has_key(fd):
            del asyncore.socket_map[fd]

    def handle_connect(self):
        pass # do nothing here, we instead wait til handle_write

    def handle_read (self):
        try:
            data = self.recv(8192*2)
        except:
            traceback.print_exc()
            raise

#        if not data:
##             self.closed_flag = 1
#             self.handle_close()

    def handle_close(self):
        self.close()

    def handle_error(self):
        traceback.print_exc()
        self.close()

    def handle_read_event (self):
        if not self.connected:
            self.handle_connect()
            self.connected = 1
            self.handle_read()
        else:
            self.handle_read()

    def handle_write_event (self):
        # getting a write implies that we are connected
        if not self.connected:
            self.handle_connect()
            self.connected = 1
        self.handle_write()

#   def __getattr__(self,k):
#        return self.__dict__[k]
            

    def readable(self):
        if not self.connected:
           return 0
        return 1

    def writable(self):
        if not len(self.data):
            return 0
        return 1

    def handle_write(self):

        try:
            sent = self.send(self.data)
            self.data = self.data[sent:]
        except socket.error, v:
            if v[0] == 32:
                return

    def close(self):

#        sys.LOG.log_debug("%s:%s%s" % \
#            (self.req_host,self.req_port,self.req_path))
            
        if self.__dict__.get('closed_flag'): return
        if 0:#self.__dict__.has_key('delay'):
#       if self.__dict__.has_key('delay'):

            self.del_channel()
            CLOSABLE.put((self.socket, time.time(), self.delay))
        else:

            self.del_channel()
            self.socket.shutdown(2)
            self.socket.close()
            global NUM_CLOSED
            NUM_CLOSED = NUM_CLOSED + 1
            if NUM_CLOSED % 50 ==0:
                print NUM_CLOSED,
                sys.stdout.flush()


class AsyncoreThread(threading.Thread):

    running = 1

    def __init__(self):
        threading.Thread.__init__(self)
        self.setName('asyncorethread')

    def stop(self):    self.running = 0        

    def run(self):

        self.start_time = time.time()

        while self.running:
            try:
                asyncore.loop(timeout=5,use_poll=0)
            except:
                traceback.print_exc()
                print 'baaahh..asyncore sucks'

            time.sleep(0.002)

        self.end_time = time.time()

        Globals.LOG.log_info("%s terminated" % self.getName()) 

